mod callbacks;

use crate::{gui::{MutexTIMManager, VRAM_HEIGHT, VRAM_WIDTH}, VRAMImgData, VRAMInfo, VRAMData};
use super::MainWindowRef;
use slint::{Model, SharedString};
use std::{cell::RefCell, ops::RangeInclusive, rc::Rc, sync::{Arc, Mutex}};
use tim_tool::logic::project::FileSettings;

pub struct NewVRAMImageInfo {
    pub image:    slint::Image,
    pub x:        u16,
    pub y:        u16,
    pub settings: FileSettings,
}

impl NewVRAMImageInfo {
    pub fn new(image: slint::Image, x: u16, y: u16, settings: FileSettings) -> NewVRAMImageInfo {
        NewVRAMImageInfo{image, x, y, settings}
    }
}

struct VRAM {
    count:     usize,
    file_list: Rc<slint::VecModel<slint::StandardListViewItem>>,
    info:      Rc<slint::VecModel<VRAMData>>
}

impl VRAM {
    pub fn push(&mut self, name: slint::StandardListViewItem, image: VRAMData) {
        self.file_list.push(name);
        self.info.push(image);
        self.count += 1;
    }

    pub fn remove(&mut self, idx: usize) {
        self.count -= 1;
        self.file_list.remove(idx);
        self.info.remove(idx);
    }

    pub fn pop(&mut self) {
        self.remove(self.count - 1);
    }
}

pub struct MainTab {
    vram: Arc<Mutex<VRAM>>,
}

pub type SharedMainTab = Rc<RefCell<MainTab>>;

impl MainTab {
    pub fn new(main_window: MainWindowRef, tim_manager: MutexTIMManager) -> SharedMainTab {
        let vram_file_list:Vec<slint::StandardListViewItem> = main_window.borrow().get_main_tab_vram_file_list().iter().collect();
        let vram_file_list                                  = Rc::new(slint::VecModel::from(vram_file_list));
        let info                                            = Rc::new(slint::VecModel::from(Vec::<VRAMData>::new()));

        main_window.borrow().set_main_tab_vram_file_list(vram_file_list.clone().into());
        main_window.borrow().set_main_tab_vram_data(info.clone().into());

        let object = SharedMainTab::new(MainTab{vram: Arc::new(Mutex::new(VRAM{count: 0, file_list: vram_file_list, info}))}.into());

        let (cloned_object, cloned_main_window, mut function) = (object.clone(), main_window.clone(), callbacks::on_move_vram_image());
        main_window.borrow().on_move_vram_image(move |idx, dx, dy| {
            function(&mut cloned_object.borrow_mut(), &cloned_main_window.borrow(), idx, dx, dy);
        });   

        let (cloned_object, cloned_main_window, mut function) = (object.clone(), main_window.clone(), callbacks::on_remove_file(tim_manager));
        main_window.borrow().on_main_tab_remove_file_clicked(move |idx| {
            function(&mut cloned_object.borrow_mut(), &cloned_main_window.borrow(), idx);
        });
        object
    }

    pub fn clear(&mut self) {
        let mut vram_data = self.vram.lock().expect("VRAM already locked");
        let     count     = vram_data.count;

        for _ in 0..count {
            vram_data.pop();
        }
    }

    pub fn add_new_vram_file(&mut self, file_name: &String, full_image: slint::Image, texture: NewVRAMImageInfo, image_palette: Option<NewVRAMImageInfo>) -> usize {
        let mut vram_data     = self.vram.lock().expect("VRAM already locked");
        let mut add_new_image = |file_name: &String, full_image: slint::Image, vram_image: NewVRAMImageInfo, palette_count: i32, is_palette: bool| {
            let encoding_str = if is_palette {"<Palette>"} else {vram_image.settings.encoding.to_str()};
            let vram_image   = VRAMData {
                images: VRAMImgData {
                    full_image, 
                    image: vram_image.image, 
                },
                info: VRAMInfo {
                    x:               vram_image.x as i32,
                    y:               vram_image.y as i32,
                    encoding_str:    SharedString::from(encoding_str),
                    use_compression: vram_image.settings.compress,
                    trans_setting:   vram_image.settings.transparency.as_idx(),
                    palette_count,
                    is_palette,
                }
            };

            vram_data.push(slint::StandardListViewItem::from(file_name.as_str()), vram_image);
        };

        let mut images_added = 1;
        add_new_image(file_name, full_image, texture, if image_palette.is_some() {1} else {0}, false);
        if let Some(image_palette) = image_palette {
            let file_name = "  => ".to_owned() + file_name.as_str() + " (Palette)";

            images_added += 1;
            add_new_image(&file_name, image_palette.image.clone(), image_palette, 0, true);
        }

        images_added
    }

    pub fn remove_vram_file(&mut self, idx: usize) -> Result<RangeInclusive<usize>, &'static str> {
        let mut vram_data = self.vram.lock().expect("VRAM already locked");
        let     extras    = {
            if let Some(element) = vram_data.info.iter().skip(idx).next() {
                if element.info.is_palette {
                    return Err("Can not remove palette. Delete image instead");
                }

                else {
                    element.info.palette_count as usize
                }
            }

            else {
                0
            }
        };

        let     range        = idx..=idx+extras;
        let mut remove_image = |idx: usize| {
            vram_data.remove(idx);
        };

        for _ in range.clone() {
            remove_image(idx);
        }
        return Ok(range);
    }

    pub fn pop_vram_files(&mut self, count: usize) {
        let mut vram_data = self.vram.lock().expect("VRAM already locked");
        for _ in 0..count {
            vram_data.pop();
        }
    }

    pub fn move_vram_image(&mut self, idx: usize, dx: i32, dy: i32) {
        let vram_data = self.vram.lock().expect("VRAM already locked");
        if let Some(mut vram_info) = vram_data.info.row_data(idx) {
            vram_info.info.x += dx;
            vram_info.info.y += dy;

            if vram_info.info.x < 0 {
                vram_info.info.x = 0;
            }

            if vram_info.info.y < 0 {
                vram_info.info.y = 0;
            }

            let (vram_img_width, vram_img_height) = (vram_info.images.image.size().width as i32, vram_info.images.image.size().height as i32);
            if (vram_info.info.x + vram_img_width) > VRAM_WIDTH as i32 {
                vram_info.info.x = VRAM_WIDTH as i32 - vram_img_width;
            }

            if (vram_info.info.y + vram_img_height) > VRAM_HEIGHT as i32 {
                vram_info.info.y = VRAM_HEIGHT as i32 - vram_img_height;
            }

            vram_data.info.set_row_data(idx, vram_info);
        }
    }

    pub fn clone_vram_info(&self) -> Vec<(String, VRAMInfo)> {
        let vram_data = self.vram.lock().expect("VRAM already locked");

        vram_data.info.iter().zip(vram_data.file_list.iter()).map(|(vram_data, file_name)| {
            (file_name.text.to_string(), vram_data.info.clone())
        }).collect()
    }
}